import time
from typing import Tuple

from rate_limiter.core.algorithms.base import RateLimitAlgorithm
from rate_limiter.core.rules import RateLimitRule, Scope
from rate_limiter.core.storage.backend import StorageBackend


class SlidingWindowAlgorithm:
    """Sliding window rate limiting algorithm using sorted sets."""

    async def allow(
        self,
        storage: StorageBackend,
        identity: str,
        rule: RateLimitRule,
        endpoint: str = "",
    ) -> Tuple[bool, int, int]:
        """
        Check if request should be allowed using sliding window.

        Args:
            storage: Storage backend
            identity: Request identity
            rule: Rate limit rule
            endpoint: Endpoint identifier

        Returns:
            Tuple of (allowed, remaining, retry_after)
        """
        # Create key based on scope
        # Include limit and window in key to ensure each rule has its own counter
        rule_id = f"{rule.limit}:{rule.window}"
        if rule.scope == Scope.GLOBAL:
            key = f"{rule.key}:{identity}:global:{rule_id}"
        elif rule.scope == Scope.ROUTER:
            key = f"{rule.key}:{identity}:router:{endpoint}:{rule_id}"
        else:  # endpoint
            key = f"{rule.key}:{identity}:endpoint:{endpoint}:{rule_id}"

        current_time = int(time.time())
        window_start = current_time - rule.window

        # Use Lua script for atomic sliding window calculation
        script = """
        local key = KEYS[1]
        local window_start = tonumber(ARGV[1])
        local limit = tonumber(ARGV[2])
        local current_time = tonumber(ARGV[3])
        local ttl = tonumber(ARGV[4])

        -- Remove old entries
        redis.call('ZREMRANGEBYSCORE', key, 0, window_start)

        -- Count current entries
        local count = redis.call('ZCARD', key)

        if count < limit then
            -- Add current request
            redis.call('ZADD', key, current_time, current_time)
            redis.call('EXPIRE', key, ttl)
            return {1, limit - count - 1, 0}
        else
            -- Get oldest entry to calculate retry_after
            local oldest = redis.call('ZRANGE', key, 0, 0, 'WITHSCORES')
            local retry_after = 0
            if oldest and #oldest > 0 then
                retry_after = tonumber(oldest[2]) + ttl - current_time
            end
            return {0, 0, retry_after}
        end
        """

        try:
            result = await storage.execute_script(
                script,
                keys=[key],
                args=[window_start, rule.limit, current_time, rule.window],
            )

            if isinstance(result, list) and len(result) >= 3:
                allowed = bool(result[0])
                remaining = int(result[1])
                retry_after = max(0, int(result[2]))
            else:
                # Fallback to simple counter if script execution fails
                return await self._fallback_allow(storage, key, rule, current_time)
        except Exception:
            # Fallback to simple counter if script execution fails
            return await self._fallback_allow(storage, key, rule, current_time)

        return allowed, remaining, retry_after

    async def _fallback_allow(
        self,
        storage: StorageBackend,
        key: str,
        rule: RateLimitRule,
        current_time: int,
    ) -> Tuple[bool, int, int]:
        """Fallback to simple counter-based approach."""
        window_key = f"{key}:window:{current_time // rule.window}"
        count = await storage.incr(window_key, rule.window)
        allowed = count <= rule.limit
        remaining = max(0, rule.limit - count)
        retry_after = rule.window - (current_time % rule.window)
        return allowed, remaining, retry_after



from typing import Protocol, Tuple

from rate_limiter.core.rules import RateLimitRule
from rate_limiter.core.storage.backend import StorageBackend


class RateLimitAlgorithm(Protocol):
    """Protocol for rate limiting algorithms."""

    async def allow(
        self,
        storage: StorageBackend,
        identity: str,
        rule: RateLimitRule,
        endpoint: str = "",
    ) -> Tuple[bool, int, int]:
        """
        Check if request should be allowed.

        Args:
            storage: Storage backend
            identity: Request identity (e.g., IP address)
            rule: Rate limit rule
            endpoint: Endpoint identifier (for scoped rules)

        Returns:
            Tuple of (allowed, remaining, retry_after)
            - allowed: True if request is allowed, False otherwise
            - remaining: Number of requests remaining in window
            - retry_after: Seconds until rate limit resets
        """
        ...



"""Rate limiting algorithms."""

from rate_limiter.core.algorithms.base import RateLimitAlgorithm
from rate_limiter.core.algorithms.fixed_window import FixedWindowAlgorithm
from rate_limiter.core.algorithms.leaky_bucket import LeakyBucketAlgorithm
from rate_limiter.core.algorithms.sliding_window import SlidingWindowAlgorithm
from rate_limiter.core.algorithms.token_bucket import TokenBucketAlgorithm

__all__ = [
    "RateLimitAlgorithm",
    "FixedWindowAlgorithm",
    "SlidingWindowAlgorithm",
    "TokenBucketAlgorithm",
    "LeakyBucketAlgorithm",
]



"""Rateon - Production-grade Python rate-limiting library."""

from rate_limiter.config import RateLimiterConfig
from rate_limiter.core.limiter import RateLimiter
from rate_limiter.exceptions import RateLimitExceeded

__version__ = "0.1.0"
__all__ = ["RateLimiter", "RateLimiterConfig", "RateLimitExceeded"]



from typing import Any, Protocol


class StorageBackend(Protocol):
    """Protocol for storage backends."""

    async def incr(self, key: str, ttl: int) -> int:
        """
        Increment counter and return new value.

        Args:
            key: Storage key
            ttl: Time to live in seconds

        Returns:
            New counter value
        """
        ...

    async def get(self, key: str) -> int:
        """
        Get current counter value.

        Args:
            key: Storage key

        Returns:
            Current counter value (0 if key doesn't exist)
        """
        ...

    async def set(self, key: str, value: int, ttl: int) -> None:
        """
        Set counter value.

        Args:
            key: Storage key
            value: Counter value
            ttl: Time to live in seconds
        """
        ...

    async def delete(self, key: str) -> None:
        """
        Delete a key.

        Args:
            key: Storage key
        """
        ...

    async def exists(self, key: str) -> bool:
        """
        Check if key exists.

        Args:
            key: Storage key

        Returns:
            True if key exists, False otherwise
        """
        ...

    async def ttl(self, key: str) -> int:
        """
        Get remaining TTL for a key.

        Args:
            key: Storage key

        Returns:
            Remaining TTL in seconds (-1 if key doesn't exist, -2 if no expiry)
        """
        ...

    async def close(self) -> None:
        """Close backend connection."""
        ...

    async def execute_script(
        self, script: str, keys: list[str], args: list[Any]
    ) -> Any:
        """
        Execute Lua script (for Redis).

        Args:
            script: Lua script content
            keys: List of keys
            args: List of arguments

        Returns:
            Script execution result
        """
        ...



import asyncio
import time
from typing import Dict, Any

from rate_limiter.core.storage.backend import StorageBackend


class MemoryStorageBackend:
    """In-memory storage backend (not production-safe)."""

    def __init__(self):
        """Initialize in-memory storage backend."""
        self._data: Dict[str, tuple[int, float]] = {}  # key -> (value, expiry)
        self._lock = asyncio.Lock()

    async def incr(self, key: str, ttl: int) -> int:
        """
        Increment counter and return new value.

        Args:
            key: Storage key
            ttl: Time to live in seconds

        Returns:
            New counter value
        """
        async with self._lock:
            self._cleanup_expired()
            current_time = time.time()
            expiry = current_time + ttl

            if key in self._data:
                value, _ = self._data[key]
                new_value = value + 1
                self._data[key] = (new_value, expiry)
            else:
                new_value = 1
                self._data[key] = (new_value, expiry)

            return new_value

    async def get(self, key: str) -> int:
        """
        Get current counter value.

        Args:
            key: Storage key

        Returns:
            Current counter value (0 if key doesn't exist or expired)
        """
        async with self._lock:
            self._cleanup_expired()
            if key in self._data:
                value, _ = self._data[key]
                return value
            return 0

    async def set(self, key: str, value: int, ttl: int) -> None:
        """
        Set counter value.

        Args:
            key: Storage key
            value: Counter value
            ttl: Time to live in seconds
        """
        async with self._lock:
            self._cleanup_expired()
            current_time = time.time()
            expiry = current_time + ttl
            self._data[key] = (value, expiry)

    async def delete(self, key: str) -> None:
        """
        Delete a key.

        Args:
            key: Storage key
        """
        async with self._lock:
            if key in self._data:
                del self._data[key]

    async def exists(self, key: str) -> bool:
        """
        Check if key exists.

        Args:
            key: Storage key

        Returns:
            True if key exists and not expired, False otherwise
        """
        async with self._lock:
            self._cleanup_expired()
            return key in self._data

    async def ttl(self, key: str) -> int:
        """
        Get remaining TTL for a key.

        Args:
            key: Storage key

        Returns:
            Remaining TTL in seconds (-1 if key doesn't exist)
        """
        async with self._lock:
            self._cleanup_expired()
            if key not in self._data:
                return -1
            _, expiry = self._data[key]
            current_time = time.time()
            remaining = int(expiry - current_time)
            return max(0, remaining)

    async def close(self) -> None:
        """Close backend connection."""
        async with self._lock:
            self._data.clear()

    async def execute_script(self, script: str, keys: list[str], args: list) -> Any:
        """
        Execute Lua script (not supported in memory backend).

        Args:
            script: Lua script content
            keys: List of keys
            args: List of arguments

        Returns:
            None (not supported)
        """
        raise NotImplementedError("Lua scripts not supported in memory backend")

    def _cleanup_expired(self) -> None:
        """Remove expired keys."""
        current_time = time.time()
        expired_keys = [
            key for key, (_, expiry) in self._data.items() if expiry <= current_time
        ]
        for key in expired_keys:
            del self._data[key]



import logging
import time
from dataclasses import dataclass
from typing import List, Optional

from rate_limiter.config import RateLimiterConfig
from rate_limiter.core.algorithms import (
    FixedWindowAlgorithm,
    LeakyBucketAlgorithm,
    RateLimitAlgorithm,
    SlidingWindowAlgorithm,
    TokenBucketAlgorithm,
)
from rate_limiter.core.identity import IdentityResolver
from rate_limiter.core.rules import Algorithm, RateLimitRule
from rate_limiter.core.storage.backend import StorageBackend
from rate_limiter.core.storage.memory import MemoryStorageBackend
from rate_limiter.core.storage.redis import RedisStorageBackend
from rate_limiter.exceptions import RateLimitExceeded, StorageBackendError

logger = logging.getLogger(__name__)


@dataclass
class RateLimitInfo:
    """Rate limit information for headers."""

    limit: int
    remaining: int
    reset: int  # Unix timestamp when rate limit resets

    def to_headers(self) -> dict[str, str]:
        """Convert to HTTP headers."""
        return {
            "X-RateLimit-Limit": str(self.limit),
            "X-RateLimit-Remaining": str(self.remaining),
            "X-RateLimit-Reset": str(self.reset),
        }


class RateLimiter:

    def __init__(
        self,
        config: Optional[RateLimiterConfig] = None,
        storage: Optional[StorageBackend] = None,
        rules: Optional[List[RateLimitRule]] = None,
    ):
        """
        Initialize rate limiter.

        Args:
            config: Rate limiter configuration
            storage: Optional storage backend (overrides config)
            rules: Optional list of default rules
        """
        self.config = config or RateLimiterConfig.from_env()
        self._storage: Optional[StorageBackend] = storage
        self.default_rules = rules or []

        # Initialize algorithms
        self._algorithms = {
            Algorithm.FIXED_WINDOW: FixedWindowAlgorithm(),
            Algorithm.SLIDING_WINDOW: SlidingWindowAlgorithm(),
            Algorithm.TOKEN_BUCKET: TokenBucketAlgorithm(),
            Algorithm.LEAKY_BUCKET: LeakyBucketAlgorithm(),
        }

        # Initialize metrics if enabled
        self._metrics = None
        if self.config.enable_metrics:
            try:
                from rate_limiter.utils.metrics import get_metrics_collector

                # Use singleton pattern to avoid duplicate metric registration
                self._metrics = get_metrics_collector()
            except ImportError:
                logger.warning("Prometheus client not available, metrics disabled")

    async def _get_storage(self) -> StorageBackend:
        """Get or create storage backend."""
        if self._storage is None:
            if self.config.backend == "redis":
                self._storage = RedisStorageBackend(
                    redis_url=self.config.redis_url,
                    key_prefix=self.config.redis_key_prefix,
                    cluster_mode=self.config.redis_cluster_mode,
                )
            else:
                self._storage = MemoryStorageBackend()
        return self._storage

    def _get_algorithm(self, algorithm: Algorithm) -> RateLimitAlgorithm:
        """Get algorithm by enum."""
        if algorithm not in self._algorithms:
            raise ValueError(f"Unknown algorithm: {algorithm}")
        return self._algorithms[algorithm]

    async def check(
        self,
        identity: str,
        rules: Optional[List[RateLimitRule]] = None,
        endpoint: str = "",
        identity_resolver: Optional[IdentityResolver] = None,
    ) -> None:
        """
        Check if request should be allowed.

        Args:
            identity: Request identity (e.g., IP address)
            rules: List of rules to check (uses default_rules if not provided)
            endpoint: Endpoint identifier (for scoped rules)
            identity_resolver: Optional identity resolver (for custom resolution)

        Raises:
            RateLimitExceeded: If rate limit is exceeded
        """
        rules_to_check = rules or self.default_rules

        if not rules_to_check:
            # No rules, allow request
            return

        # Sort rules by priority (higher priority first)
        sorted_rules = sorted(rules_to_check, key=lambda r: r.priority, reverse=True)

        storage = await self._get_storage()

        for rule in sorted_rules:
            try:
                algorithm_impl = self._get_algorithm(rule.algorithm)
                allowed, remaining, retry_after = await algorithm_impl.allow(
                    storage, identity, rule, endpoint
                )

                # Record metrics
                if self._metrics:
                    self._metrics.record_request(rule.key, allowed)

                # Log request
                if self.config.enable_logging:
                    logger.info(
                        f"Rate limit check: identity={identity}, rule={rule.key}, "
                        f"allowed={allowed}, remaining={remaining}, retry_after={retry_after}"
                    )

                if not allowed:
                    raise RateLimitExceeded(
                        retry_after=retry_after,
                        limit=rule.limit,
                        window=rule.window,
                    )
            except StorageBackendError as e:
                logger.error(f"Storage backend error: {e}")
                if self.config.fail_closed:
                    # Fail closed: deny request on backend failure
                    raise RateLimitExceeded(
                        retry_after=60,
                        limit=0,
                        window=60,
                        message="Rate limiter backend unavailable",
                    )
                # Fail open: allow request on backend failure
                logger.warning("Allowing request due to backend failure (fail_open mode)")

    async def check_with_info(
        self,
        identity: str,
        rules: Optional[List[RateLimitRule]] = None,
        endpoint: str = "",
        identity_resolver: Optional[IdentityResolver] = None,
    ) -> tuple[Optional[RateLimitInfo], Optional[RateLimitExceeded]]:
        """
        Check rate limit and return info. Returns (info, None) if allowed, (None, exception) if not.

        Args:
            identity: Request identity (e.g., IP address)
            rules: List of rules to check (uses default_rules if not provided)
            endpoint: Endpoint identifier (for scoped rules)
            identity_resolver: Optional identity resolver (for custom resolution)

        Returns:
            Tuple of (RateLimitInfo or None, RateLimitExceeded or None)
        """
        rules_to_check = rules or self.default_rules

        if not rules_to_check:
            return None, None

        # Sort rules by priority (higher priority first)
        sorted_rules = sorted(rules_to_check, key=lambda r: r.priority, reverse=True)

        storage = await self._get_storage()
        current_time = int(time.time())

        for rule in sorted_rules:
            try:
                algorithm_impl = self._get_algorithm(rule.algorithm)
                allowed, remaining, retry_after = await algorithm_impl.allow(
                    storage, identity, rule, endpoint
                )

                # Calculate reset timestamp
                reset_timestamp = current_time + retry_after
                rate_limit_info = RateLimitInfo(
                    limit=rule.limit,
                    remaining=remaining,
                    reset=reset_timestamp,
                )

                # Record metrics
                if self._metrics:
                    self._metrics.record_request(rule.key, allowed)

                # Log request
                if self.config.enable_logging:
                    logger.info(
                        f"Rate limit check: identity={identity}, rule={rule.key}, "
                        f"allowed={allowed}, remaining={remaining}, retry_after={retry_after}"
                    )

                if not allowed:
                    exception = RateLimitExceeded(
                        retry_after=retry_after,
                        limit=rule.limit,
                        window=rule.window,
                    )
                    return None, exception

                # Return info for the first (highest priority) rule that passes
                return rate_limit_info, None

            except StorageBackendError as e:
                logger.error(f"Storage backend error: {e}")
                if self.config.fail_closed:
                    # Fail closed: deny request on backend failure
                    exception = RateLimitExceeded(
                        retry_after=60,
                        limit=0,
                        window=60,
                        message="Rate limiter backend unavailable",
                    )
                    return None, exception
                # Fail open: allow request on backend failure
                logger.warning("Allowing request due to backend failure (fail_open mode)")
                continue

        return None, None

    async def get_rate_limit_info(
        self,
        identity: str,
        rules: Optional[List[RateLimitRule]] = None,
        endpoint: str = "",
    ) -> Optional[RateLimitInfo]:
        """
        Get rate limit information without raising exceptions.
        Note: This calls check_with_info but doesn't increment counters.

        Args:
            identity: Request identity (e.g., IP address)
            rules: List of rules to check (uses default_rules if not provided)
            endpoint: Endpoint identifier (for scoped rules)

        Returns:
            RateLimitInfo if rules exist, None otherwise
        """
        # Use check_with_info but ignore the exception part
        info, _ = await self.check_with_info(identity=identity, rules=rules, endpoint=endpoint)
        return info

    async def close(self) -> None:
        """Close rate limiter and cleanup resources."""
        if self._storage:
            await self._storage.close()
            self._storage = None



from typing import Callable, List, Optional

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from rate_limiter.config import RateLimiterConfig
from rate_limiter.core.identity import IdentityResolver, get_identity_resolver
from rate_limiter.core.limiter import RateLimiter
from rate_limiter.core.rules import RateLimitRule
from rate_limiter.exceptions import RateLimitExceeded


class RateLimiterMiddleware(BaseHTTPMiddleware):
    """Starlette middleware for rate limiting."""

    def __init__(
        self,
        app: ASGIApp,
        rules: Optional[List[RateLimitRule]] = None,
        config: Optional[RateLimiterConfig] = None,
        identity_resolver: Optional[IdentityResolver] = None,
    ):
        """
        Initialize rate limiter middleware.

        Args:
            app: ASGI application
            rules: List of rate limit rules
            config: Rate limiter configuration
            identity_resolver: Optional custom identity resolver
        """
        super().__init__(app)
        self.config = config or RateLimiterConfig.from_env()
        self.limiter = RateLimiter(config=self.config, rules=rules)
        self.identity_resolver = identity_resolver or get_identity_resolver(
            "ip", trust_proxy=self.config.trust_proxy_headers
        )

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        """
        Process request through rate limiter.

        Args:
            request: Starlette request
            call_next: Next middleware/handler

        Returns:
            Response
        """
        # Resolve identity
        identity = await self.identity_resolver.resolve(request)

        # Get endpoint path for scoped rules
        endpoint = request.url.path

        # Check rate limits and get info in one call
        rate_limit_info, rate_limit_exception = await self.limiter.check_with_info(
            identity=identity, endpoint=endpoint
        )

        if rate_limit_exception:
            # Return 429 Too Many Requests with rate limit headers
            e = rate_limit_exception
            import time

            reset_timestamp = int(time.time()) + e.retry_after
            headers = {
                "Retry-After": str(e.retry_after),
                "X-RateLimit-Limit": str(e.limit),
                "X-RateLimit-Remaining": "0",
                "X-RateLimit-Reset": str(reset_timestamp),
            }
            return Response(
                content=f"Rate limit exceeded. Retry after {e.retry_after} seconds.",
                status_code=429,
                headers=headers,
            )

        # Process request
        response = await call_next(request)

        # Add rate limit headers to successful response
        if rate_limit_info:
            headers = rate_limit_info.to_headers()
            for key, value in headers.items():
                response.headers[key] = value

        return response



import time
from typing import Tuple

from rate_limiter.core.algorithms.base import RateLimitAlgorithm
from rate_limiter.core.rules import RateLimitRule, Scope
from rate_limiter.core.storage.backend import StorageBackend


class LeakyBucketAlgorithm:
    """Leaky bucket rate limiting algorithm."""

    async def allow(
        self,
        storage: StorageBackend,
        identity: str,
        rule: RateLimitRule,
        endpoint: str = "",
    ) -> Tuple[bool, int, int]:
        """
        Check if request should be allowed using leaky bucket.

        Args:
            storage: Storage backend
            identity: Request identity
            rule: Rate limit rule
            endpoint: Endpoint identifier

        Returns:
            Tuple of (allowed, remaining, retry_after)
        """
        # Create key based on scope
        # Include limit and window in key to ensure each rule has its own counter
        rule_id = f"{rule.limit}:{rule.window}"
        if rule.scope == Scope.GLOBAL:
            key = f"{rule.key}:{identity}:global:{rule_id}"
        elif rule.scope == Scope.ROUTER:
            key = f"{rule.key}:{identity}:router:{endpoint}:{rule_id}"
        else:  # endpoint
            key = f"{rule.key}:{identity}:endpoint:{endpoint}:{rule_id}"

        current_time = int(time.time())

        # Leaky bucket parameters
        # Capacity = limit (bucket size)
        # Leak rate = limit / window (requests per second)
        capacity = rule.limit
        leak_rate = rule.limit / rule.window

        # Use Lua script for atomic leaky bucket operations
        script = """
        local key = KEYS[1]
        local current_time = tonumber(ARGV[1])
        local capacity = tonumber(ARGV[2])
        local leak_rate = tonumber(ARGV[3])
        local ttl = tonumber(ARGV[4])

        local bucket = redis.call('HMGET', key, 'level', 'last_leak')
        local level = tonumber(bucket[1]) or 0
        local last_leak = tonumber(bucket[2]) or current_time

        -- Leak water based on time elapsed
        local time_elapsed = current_time - last_leak
        local leaked = math.floor(time_elapsed * leak_rate)
        level = math.max(0, level - leaked)
        last_leak = current_time

        -- Check if bucket has space
        if level < capacity then
            level = level + 1
            redis.call('HMSET', key, 'level', level, 'last_leak', last_leak)
            redis.call('EXPIRE', key, ttl)
            local retry_after = 0
            if level >= capacity then
                retry_after = math.ceil((level - capacity + 1) / leak_rate)
            end
            return {1, capacity - level, retry_after}
        else
            -- Bucket is full, calculate retry_after
            local retry_after = math.ceil((level - capacity + 1) / leak_rate)
            redis.call('HMSET', key, 'level', level, 'last_leak', last_leak)
            redis.call('EXPIRE', key, ttl)
            return {0, 0, retry_after}
        end
        """

        try:
            result = await storage.execute_script(
                script,
                keys=[key],
                args=[current_time, capacity, leak_rate, rule.window],
            )

            if isinstance(result, list) and len(result) >= 3:
                allowed = bool(result[0])
                remaining = int(result[1])
                retry_after = max(0, int(result[2]))
            else:
                # Fallback to simple counter
                return await self._fallback_allow(storage, key, rule, current_time)
        except Exception:
            # Fallback to simple counter
            return await self._fallback_allow(storage, key, rule, current_time)

        return allowed, remaining, retry_after

    async def _fallback_allow(
        self,
        storage: StorageBackend,
        key: str,
        rule: RateLimitRule,
        current_time: int,
    ) -> Tuple[bool, int, int]:
        """Fallback to simple counter-based approach."""
        window_key = f"{key}:window:{current_time // rule.window}"
        count = await storage.incr(window_key, rule.window)
        allowed = count <= rule.limit
        remaining = max(0, rule.limit - count)
        retry_after = rule.window - (current_time % rule.window)
        return allowed, remaining, retry_after



"""FastAPI integrations."""

from rate_limiter.integrations.fastapi.decorators import rate_limit
from rate_limiter.integrations.fastapi.dependencies import rate_limit_dep
from rate_limiter.integrations.fastapi.middleware import RateLimiterMiddleware

__all__ = ["RateLimiterMiddleware", "rate_limit", "rate_limit_dep"]



from typing import Optional

try:
    from prometheus_client import Counter, Histogram, CollectorRegistry, REGISTRY
except ImportError:
    Counter = None
    Histogram = None
    CollectorRegistry = None
    REGISTRY = None

# Global metrics instance to avoid duplicate registration
_metrics_instance: Optional["MetricsCollector"] = None


class MetricsCollector:
    """Collects Prometheus metrics for rate limiting."""

    def __init__(self, registry: Optional[CollectorRegistry] = None):
        """
        Initialize metrics collector.
        
        Args:
            registry: Optional Prometheus registry. If None, uses default registry.
                     Use a custom registry in tests to avoid conflicts.
        """
        if Counter is None or Histogram is None:
            raise ImportError("prometheus_client not installed")

        self.registry = registry or REGISTRY

        self.requests_total = Counter(
            "rate_limiter_requests_total",
            "Total rate limiter requests",
            ["rule_key", "status"],
            registry=self.registry,
        )

        self.requests_allowed = Counter(
            "rate_limiter_requests_allowed",
            "Allowed rate limiter requests",
            ["rule_key"],
            registry=self.registry,
        )

        self.requests_blocked = Counter(
            "rate_limiter_requests_blocked",
            "Blocked rate limiter requests",
            ["rule_key"],
            registry=self.registry,
        )

        self.request_duration = Histogram(
            "rate_limiter_request_duration_seconds",
            "Rate limiter request duration",
            ["rule_key"],
            registry=self.registry,
        )

    def record_request(self, rule_key: str, allowed: bool) -> None:
        """
        Record a rate limit check.

        Args:
            rule_key: Rule key identifier
            allowed: Whether request was allowed
        """
        status = "allowed" if allowed else "blocked"
        self.requests_total.labels(rule_key=rule_key, status=status).inc()

        if allowed:
            self.requests_allowed.labels(rule_key=rule_key).inc()
        else:
            self.requests_blocked.labels(rule_key=rule_key).inc()


def get_metrics_collector(registry: Optional[CollectorRegistry] = None) -> MetricsCollector:
    """
    Get or create a metrics collector instance.
    
    Uses singleton pattern with default registry to avoid duplicates.
    For tests, pass a custom registry.
    
    Args:
        registry: Optional custom registry (useful for tests)
        
    Returns:
        MetricsCollector instance
    """
    global _metrics_instance
    
    if registry is None:
        # Use singleton for default registry
        if _metrics_instance is None:
            _metrics_instance = MetricsCollector()
        return _metrics_instance
    else:
        # Create new instance for custom registry
        return MetricsCollector(registry=registry)



from typing import Optional


class RateLimitExceeded(Exception):
    """Raised when a rate limit is exceeded."""

    def __init__(
        self,
        retry_after: int,
        limit: int,
        window: int,
        message: Optional[str] = None,
    ):
        """
        Initialize rate limit exceeded exception.

        Args:
            retry_after: Seconds until the rate limit resets
            limit: The rate limit that was exceeded
            window: The time window in seconds
            message: Optional custom error message
        """
        self.retry_after = retry_after
        self.limit = limit
        self.window = window
        self.message = message or f"Rate limit exceeded: {limit} requests per {window} seconds"
        super().__init__(self.message)

    def __repr__(self) -> str:
        return (
            f"RateLimitExceeded(retry_after={self.retry_after}, "
            f"limit={self.limit}, window={self.window})"
        )


class RateLimiterError(Exception):
    """Base exception for rate limiter errors."""

    pass


class StorageBackendError(RateLimiterError):
    """Raised when storage backend operations fail."""

    pass


class ConfigurationError(RateLimiterError):
    """Raised when configuration is invalid."""

    pass



from dataclasses import dataclass
from enum import Enum


class Algorithm(str, Enum):
    FIXED_WINDOW = "fixed_window"
    SLIDING_WINDOW = "sliding_window"
    TOKEN_BUCKET = "token_bucket"
    LEAKY_BUCKET = "leaky_bucket"


class Scope(str, Enum):
    ENDPOINT = "endpoint"
    ROUTER = "router"
    GLOBAL = "global"


@dataclass(frozen=True)
class RateLimitRule:
    """
    Defines a rate limit rule.

    Attributes:
        key: Identity key type ("ip", "user_id", "api_key", or custom)
        limit: Maximum number of requests allowed
        window: Time window in seconds
        algorithm: Algorithm to use
        scope: Scope of the rule
        priority: Rule priority (higher = checked first)
    """

    key: str
    limit: int
    window: int
    algorithm: Algorithm = Algorithm.SLIDING_WINDOW
    scope: Scope = Scope.ENDPOINT
    priority: int = 0

    def __post_init__(self) -> None:
        """Validate rule parameters."""
        if self.limit <= 0:
            raise ValueError("limit must be positive")
        if self.window <= 0:
            raise ValueError("window must be positive")

    @classmethod
    def from_string(cls, rule_str: str, key: str = "ip") -> "RateLimitRule":
        """
        Parse rate limit rule from string format.

        Examples:
            "100/minute" -> RateLimitRule(limit=100, window=60)
            "10/second" -> RateLimitRule(limit=10, window=1)
            "5/hour" -> RateLimitRule(limit=5, window=3600)

        Args:
            rule_str: Rule string in format "limit/unit"
            key: Identity key type

        Returns:
            RateLimitRule instance
        """
        parts = rule_str.split("/")
        if len(parts) != 2:
            raise ValueError(f"Invalid rule format: {rule_str}. Expected 'limit/unit'")

        try:
            limit = int(parts[0])
        except ValueError:
            raise ValueError(f"Invalid limit: {parts[0]}")

        unit = parts[1].lower()
        unit_map = {
            "second": 1,
            "minute": 60,
            "hour": 3600,
            "day": 86400,
        }

        if unit not in unit_map:
            raise ValueError(f"Invalid unit: {unit}. Must be one of: {list(unit_map.keys())}")

        window = unit_map[unit]

        return cls(key=key, limit=limit, window=window)



import hashlib
from typing import Any, Awaitable, Dict, Optional, Union, cast

import redis.asyncio as aioredis
from redis.asyncio.cluster import RedisCluster

from rate_limiter.exceptions import StorageBackendError


class RedisStorageBackend:
    """Redis storage backend with atomic operations. Supports both standalone and cluster mode."""

    def __init__(self, redis_url: str, key_prefix: str = "rateon", cluster_mode: bool = False):
        """
        Initialize Redis storage backend.

        Args:
            redis_url: Redis connection URL. For cluster mode, provide a single node URL
                      (client will auto-discover other nodes). Multiple URLs (comma-separated)
                      are supported for redundancy but only one is required.
            key_prefix: Prefix for all Redis keys
            cluster_mode: Enable Redis Cluster mode
        """
        self.redis_url = redis_url
        self.key_prefix = key_prefix
        self.cluster_mode = cluster_mode
        self._client: Optional[Union[aioredis.Redis, RedisCluster]] = None
        self._scripts: Dict[str, str] = {}  # Cache for Lua scripts

    async def _get_client(self) -> Union[aioredis.Redis, RedisCluster]:
        """Get or create Redis client (standalone or cluster)."""
        if self._client is None:
            try:
                if self.cluster_mode:
                    # Parse startup node(s) from URL
                    # Only one node is needed - client will discover the rest of the cluster
                    nodes = self._parse_cluster_nodes(self.redis_url)
                    self._client = RedisCluster(
                        startup_nodes=nodes,
                        decode_responses=False,  # We'll handle encoding ourselves
                        socket_connect_timeout=1,
                        socket_timeout=1,
                        skip_full_coverage_check=True,  # Allow partial cluster coverage
                    )
                    # Initialize cluster connection (discovers all nodes automatically)
                    await self._client.initialize()
                else:
                    self._client = await aioredis.from_url(
                        self.redis_url,
                        decode_responses=False,  # We'll handle encoding ourselves
                        socket_connect_timeout=1,
                        socket_timeout=1,
                    )
            except Exception as e:
                raise StorageBackendError(f"Failed to connect to Redis: {e}") from e
        return self._client

    def _parse_cluster_nodes(self, url_or_nodes: str) -> list[dict]:
        """
        Parse cluster startup node(s) from URL string.

        For Redis Cluster, only one node is required - the client will automatically
        discover all other nodes in the cluster. Multiple nodes can be provided for
        redundancy (if one is down, others can be tried).

        Args:
            url_or_nodes: Single Redis URL or comma-separated URLs for redundancy

        Returns:
            List of node dictionaries with 'host' and 'port' keys
        """
        nodes = []
        # Split by comma if multiple URLs provided (for redundancy)
        node_strings = url_or_nodes.split(",") if "," in url_or_nodes else [url_or_nodes]
        
        for node_str in node_strings:
            node_str = node_str.strip()
            if not node_str:
                continue

            # Parse URL format: redis://host:port or host:port
            if node_str.startswith("redis://"):
                node_str = node_str[8:]
            elif node_str.startswith("rediss://"):
                node_str = node_str[9:]

            if ":" in node_str:
                host, port = node_str.rsplit(":", 1)
                nodes.append({"host": host, "port": int(port)})
            else:
                # Default port
                nodes.append({"host": node_str, "port": 6379})

        if not nodes:
            raise ValueError("No valid cluster nodes found. Provide at least one node URL.")

        return nodes

    def _make_key(self, key: str) -> bytes:
        """
        Create Redis key with prefix and sanitization.

        For cluster mode, uses hash tags to ensure keys are in the same slot.

        Args:
            key: Original key

        Returns:
            Sanitized Redis key as bytes
        """
        # Sanitize key to prevent injection
        safe_key = hashlib.sha256(key.encode()).hexdigest()[:32]
        
        if self.cluster_mode:
            # Use hash tags for cluster mode to ensure keys are in the same slot
            # This is critical for Lua scripts which require all keys in the same slot
            full_key = f"{self.key_prefix}:{{{safe_key}}}"
        else:
            full_key = f"{self.key_prefix}:{safe_key}"
        
        return full_key.encode()

    async def incr(self, key: str, ttl: int) -> int:
        """
        Increment counter atomically and return new value.

        Args:
            key: Storage key
            ttl: Time to live in seconds

        Returns:
            New counter value
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)

            # Use Lua script for atomic increment + TTL
            script = """
            local current = redis.call('INCR', KEYS[1])
            if current == 1 then
                redis.call('EXPIRE', KEYS[1], ARGV[1])
            end
            return current
            """
            # Handle both standalone and cluster modes
            # Type cast needed because redis-py type stubs may not correctly infer awaitable return type
            eval_result = client.eval(script, 1, redis_key, ttl)
            result = await cast(Awaitable[Any], eval_result)
            return int(result) if result else 0
        except Exception as e:
            raise StorageBackendError(f"Redis INCR failed: {e}") from e

    async def get(self, key: str) -> int:
        """
        Get current counter value.

        Args:
            key: Storage key

        Returns:
            Current counter value (0 if key doesn't exist)
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)
            result = await client.get(redis_key)
            return int(result) if result else 0
        except Exception as e:
            raise StorageBackendError(f"Redis GET failed: {e}") from e

    async def set(self, key: str, value: int, ttl: int) -> None:
        """
        Set counter value.

        Args:
            key: Storage key
            value: Counter value
            ttl: Time to live in seconds
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)
            await client.setex(redis_key, ttl, value)
        except Exception as e:
            raise StorageBackendError(f"Redis SET failed: {e}") from e

    async def delete(self, key: str) -> None:
        """
        Delete a key.

        Args:
            key: Storage key
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)
            await client.delete(redis_key)
        except Exception as e:
            raise StorageBackendError(f"Redis DELETE failed: {e}") from e

    async def exists(self, key: str) -> bool:
        """
        Check if key exists.

        Args:
            key: Storage key

        Returns:
            True if key exists, False otherwise
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)
            result = await client.exists(redis_key)
            return bool(result)
        except Exception as e:
            raise StorageBackendError(f"Redis EXISTS failed: {e}") from e

    async def ttl(self, key: str) -> int:
        """
        Get remaining TTL for a key.

        Args:
            key: Storage key

        Returns:
            Remaining TTL in seconds (-1 if key doesn't exist, -2 if no expiry)
        """
        try:
            client = await self._get_client()
            redis_key = self._make_key(key)
            result = await client.ttl(redis_key)
            return int(result) if result is not None else -1
        except Exception as e:
            raise StorageBackendError(f"Redis TTL failed: {e}") from e

    async def close(self) -> None:
        """Close Redis connection."""
        if self._client:
            await self._client.aclose()
            self._client = None

    async def execute_script(self, script: str, keys: list[str], args: list[Any]) -> Any:
        """
        Execute Lua script with keys and arguments.

        For cluster mode, ensures all keys use hash tags to be in the same slot.

        Args:
            script: Lua script content
            keys: List of keys
            args: List of arguments

        Returns:
            Script execution result
        """
        try:
            client = await self._get_client()
            # Hash script to cache it
            script_hash = hashlib.sha1(script.encode()).hexdigest()

            # Convert keys to bytes (hash tags are added in _make_key for cluster mode)
            redis_keys = [self._make_key(k) for k in keys]

            # Convert args to strings/bytes
            redis_args = [str(arg).encode() if not isinstance(arg, bytes) else arg for arg in args]

            if self.cluster_mode:
                # For cluster mode, use eval with script content directly
                # EVALSHA doesn't work well with cluster as scripts need to be loaded on each node
                eval_result = client.eval(script, len(redis_keys), *redis_keys, *redis_args)
                result = await cast(Awaitable[Any], eval_result)
            else:
                # Use EVALSHA if script is cached, otherwise use EVAL
                try:
                    evalsha_result = client.evalsha(
                        script_hash, len(redis_keys), *redis_keys, *redis_args
                    )
                    result = await cast(Awaitable[Any], evalsha_result)
                except aioredis.ResponseError:
                    # Script not cached, use EVAL and cache it
                    eval_result = client.eval(script, len(redis_keys), *redis_keys, *redis_args)
                    result = await cast(Awaitable[Any], eval_result)
                    # Cache script for future use
                    await client.script_load(script)
                    self._scripts[script_hash] = script

            return result
        except Exception as e:
            raise StorageBackendError(f"Redis script execution failed: {e}") from e



from typing import Callable, Optional

from fastapi import Depends, Request

from rate_limiter.config import RateLimiterConfig
from rate_limiter.core.identity import IdentityResolver, get_identity_resolver
from rate_limiter.core.limiter import RateLimiter
from rate_limiter.core.rules import RateLimitRule
from rate_limiter.exceptions import RateLimitExceeded


def rate_limit_dep(
    rule_str: str,
    key: str = "ip",
    config: Optional[RateLimiterConfig] = None,
    identity_resolver: Optional[IdentityResolver] = None,
) -> Callable:
    """
    Create a FastAPI dependency for rate limiting.

    Args:
        rule_str: Rate limit rule string (e.g., "10/minute")
        key: Identity key type ("ip", "user_id", "api_key")
        config: Optional rate limiter configuration
        identity_resolver: Optional custom identity resolver

    Returns:
        Dependency function

    Example:
        router = APIRouter(
            dependencies=[Depends(rate_limit_dep("10/minute", key="ip"))]
        )
    """
    rule = RateLimitRule.from_string(rule_str, key=key)
    limiter_config = config or RateLimiterConfig.from_env()
    limiter = RateLimiter(config=limiter_config, rules=[rule])
    resolver = identity_resolver or get_identity_resolver(
        key, trust_proxy=limiter_config.trust_proxy_headers
    )

    async def dependency(request: Request) -> None:
        """Rate limit dependency."""
        # Resolve identity
        identity = await resolver.resolve(request)

        # Get endpoint path
        endpoint = request.url.path if hasattr(request, "url") else ""

        # Check rate limit
        try:
            await limiter.check(identity=identity, endpoint=endpoint)
        except RateLimitExceeded as e:
            from fastapi import HTTPException

            raise HTTPException(
                status_code=429,
                detail=f"Rate limit exceeded. Retry after {e.retry_after} seconds.",
                headers={"Retry-After": str(e.retry_after)},
            )

    return dependency



import hashlib
from typing import Any, Callable, Optional, Protocol, cast


class IdentityResolver(Protocol):
    """Protocol for identity resolvers."""

    async def resolve(self, context: Any) -> str:
        """
        Resolve identity from context.

        Args:
            context: Framework-specific request context

        Returns:
            Identity string (e.g., IP address, user ID)
        """
        ...


class IPIdentityResolver:
    """Resolves identity from IP address."""

    def __init__(self, trust_proxy: bool = False):
        """
        Initialize IP identity resolver.

        Args:
            trust_proxy: Whether to trust X-Forwarded-For headers
        """
        self.trust_proxy = trust_proxy

    async def resolve(self, request: Any) -> str:
        """
        Resolve IP address from request.

        Args:
            request: Framework request object

        Returns:
            IP address string
        """
        # Try to get IP from request object
        if hasattr(request, "client") and request.client:
            ip = request.client.host
        elif hasattr(request, "scope") and "client" in request.scope:
            ip = request.scope.get("client", ("", 0))[0]
        else:
            ip = "unknown"

        # Check X-Forwarded-For if trusted
        if self.trust_proxy:
            forwarded_for = self._get_header(request, "X-Forwarded-For")
            if forwarded_for:
                # Take the first IP (original client)
                ip = forwarded_for.split(",")[0].strip()

        # Fallback to X-Real-IP
        if ip == "unknown" or not ip:
            real_ip = self._get_header(request, "X-Real-IP")
            if real_ip:
                ip = real_ip.strip()

        return ip or "unknown"

    def _get_header(self, request: Any, header_name: str) -> Optional[str]:
        """Get header value from request."""
        if hasattr(request, "headers"):
            if isinstance(request.headers, dict):
                return request.headers.get(header_name)
            elif hasattr(request.headers, "get"):
                return request.headers.get(header_name)
        return None


class UserIdentityResolver:
    """Resolves identity from user ID in request."""

    def __init__(self, user_id_extractor: Optional[Callable[[Any], str]] = None):
        """
        Initialize user identity resolver.

        Args:
            user_id_extractor: Optional function to extract user ID from request
        """
        self.user_id_extractor = user_id_extractor

    async def resolve(self, request: Any) -> str:
        """
        Resolve user ID from request.

        Args:
            request: Framework request object

        Returns:
            User ID string
        """
        if self.user_id_extractor:
            return self.user_id_extractor(request)

        # Try common patterns
        if hasattr(request, "user") and request.user:
            if hasattr(request.user, "id"):
                return str(request.user.id)
            if hasattr(request.user, "user_id"):
                return str(request.user.user_id)

        if hasattr(request, "state") and hasattr(request.state, "user"):
            user = request.state.user
            if hasattr(user, "id"):
                return str(user.id)

        # Try JWT sub claim
        if hasattr(request, "state") and hasattr(request.state, "jwt_payload"):
            payload = request.state.jwt_payload
            if isinstance(payload, dict) and "sub" in payload:
                return str(payload["sub"])

        return "anonymous"


class APIKeyIdentityResolver:
    """Resolves identity from API key in request."""

    def __init__(self, header_name: str = "X-API-Key"):
        """
        Initialize API key identity resolver.

        Args:
            header_name: Header name containing API key
        """
        self.header_name = header_name

    async def resolve(self, request: Any) -> str:
        """
        Resolve API key from request.

        Args:
            request: Framework request object

        Returns:
            API key string (hashed for security)
        """
        api_key = self._get_header(request)
        if not api_key:
            return "no_key"

        # Hash the API key to avoid storing raw keys in Redis
        return self._hash_key(api_key)

    def _get_header(self, request: Any) -> Optional[str]:
        """Get header value from request."""
        if hasattr(request, "headers"):
            if isinstance(request.headers, dict):
                return request.headers.get(self.header_name)
            elif hasattr(request.headers, "get"):
                return request.headers.get(self.header_name)
        return None

    def _hash_key(self, key: str) -> str:
        """Hash API key for safe storage."""
        # Use HMAC with a constant secret to ensure consistent hashing
        # In production, you might want to use a configurable secret
        return hashlib.sha256(key.encode()).hexdigest()[:16]


class CompositeIdentityResolver:
    """Resolves identity from multiple sources."""

    def __init__(self, resolvers: list[tuple[str, IdentityResolver]]):
        """
        Initialize composite identity resolver.

        Args:
            resolvers: List of (key, resolver) tuples
        """
        self.resolvers = resolvers

    async def resolve(self, request: Any) -> str:
        """
        Resolve composite identity from request.

        Args:
            request: Framework request object

        Returns:
            Composite identity string (e.g., "ip:user_id")
        """
        parts = []
        for key, resolver in self.resolvers:
            identity = await resolver.resolve(request)
            parts.append(f"{key}:{identity}")
        return ":".join(parts)


def get_identity_resolver(key: str, trust_proxy: bool = False) -> IdentityResolver:
    """
    Get identity resolver for key type.

    Args:
        key: Key type ("ip", "user_id", "api_key")
        trust_proxy: Whether to trust proxy headers

    Returns:
        IdentityResolver instance
    """
    if key == "ip":
        return cast(IdentityResolver, IPIdentityResolver(trust_proxy=trust_proxy))
    elif key == "user_id":
        return cast(IdentityResolver, UserIdentityResolver())
    elif key == "api_key":
        return cast(IdentityResolver, APIKeyIdentityResolver())
    else:
        raise ValueError(f"Unknown key type: {key}")



import os
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class RateLimiterConfig:
    """Configuration for rate limiter."""

    backend: str = "redis"
    redis_url: str = "redis://localhost:6379"
    redis_cluster_mode: bool = False  # Enable Redis Cluster mode
    redis_key_prefix: str = "rateon"
    default_limits: List[str] = field(default_factory=list)
    trust_proxy_headers: bool = False
    fail_closed: bool = True  # Deny on backend failure
    enable_metrics: bool = True
    enable_logging: bool = True
    log_level: str = "INFO"

    @classmethod
    def from_env(cls) -> "RateLimiterConfig":
        """
        Create configuration from environment variables.

        Environment variables:
            RATE_LIMITER_BACKEND: Storage backend ("redis" or "memory")
            RATE_LIMITER_REDIS_URL: Redis connection URL. For cluster mode, a single node URL
                                   is sufficient (client auto-discovers other nodes). Multiple
                                   URLs (comma-separated) are optional for redundancy.
            RATE_LIMITER_REDIS_CLUSTER_MODE: Enable Redis Cluster mode (true/false)
            RATE_LIMITER_REDIS_KEY_PREFIX: Prefix for Redis keys
            RATE_LIMITER_TRUST_PROXY_HEADERS: Trust X-Forwarded-For headers (true/false)
            RATE_LIMITER_FAIL_CLOSED: Deny on backend failure (true/false)
            RATE_LIMITER_ENABLE_METRICS: Enable Prometheus metrics (true/false)
            RATE_LIMITER_ENABLE_LOGGING: Enable structured logging (true/false)
            RATE_LIMITER_LOG_LEVEL: Logging level (DEBUG, INFO, WARNING, ERROR)
        """
        backend = os.getenv("RATE_LIMITER_BACKEND", "redis")
        redis_url = os.getenv("RATE_LIMITER_REDIS_URL", "redis://localhost:6379")
        redis_cluster_mode = os.getenv("RATE_LIMITER_REDIS_CLUSTER_MODE", "false").lower() == "true"
        redis_key_prefix = os.getenv("RATE_LIMITER_REDIS_KEY_PREFIX", "rateon")
        trust_proxy_headers = os.getenv("RATE_LIMITER_TRUST_PROXY_HEADERS", "false").lower() == "true"
        fail_closed = os.getenv("RATE_LIMITER_FAIL_CLOSED", "true").lower() == "true"
        enable_metrics = os.getenv("RATE_LIMITER_ENABLE_METRICS", "true").lower() == "true"
        enable_logging = os.getenv("RATE_LIMITER_ENABLE_LOGGING", "true").lower() == "true"
        log_level = os.getenv("RATE_LIMITER_LOG_LEVEL", "INFO")

        return cls(
            backend=backend,
            redis_url=redis_url,
            redis_cluster_mode=redis_cluster_mode,
            redis_key_prefix=redis_key_prefix,
            trust_proxy_headers=trust_proxy_headers,
            fail_closed=fail_closed,
            enable_metrics=enable_metrics,
            enable_logging=enable_logging,
            log_level=log_level,
        )

    def __post_init__(self) -> None:
        """Validate configuration."""
        if self.backend not in ("redis", "memory"):
            raise ValueError(f"Invalid backend: {self.backend}. Must be 'redis' or 'memory'")



import time
from typing import Tuple

from rate_limiter.core.algorithms.base import RateLimitAlgorithm
from rate_limiter.core.rules import RateLimitRule, Scope
from rate_limiter.core.storage.backend import StorageBackend


class FixedWindowAlgorithm:
    """Fixed window rate limiting algorithm."""

    async def allow(
        self,
        storage: StorageBackend,
        identity: str,
        rule: RateLimitRule,
        endpoint: str = "",
    ) -> Tuple[bool, int, int]:
        """
        Check if request should be allowed using fixed window.

        Args:
            storage: Storage backend
            identity: Request identity
            rule: Rate limit rule
            endpoint: Endpoint identifier

        Returns:
            Tuple of (allowed, remaining, retry_after)
        """
        # Create key based on scope
        # Include limit and window in key to ensure each rule has its own counter
        rule_id = f"{rule.limit}:{rule.window}"
        if rule.scope == Scope.GLOBAL:
            key = f"{rule.key}:{identity}:global:{rule_id}"
        elif rule.scope == Scope.ROUTER:
            key = f"{rule.key}:{identity}:router:{endpoint}:{rule_id}"
        else:  # endpoint
            key = f"{rule.key}:{identity}:endpoint:{endpoint}:{rule_id}"

        # Use window number as part of key to create fixed windows
        current_time = int(time.time())
        window_number = current_time // rule.window
        window_key = f"{key}:window:{window_number}"

        # Increment counter
        count = await storage.incr(window_key, rule.window)

        # Check if limit exceeded
        allowed = count <= rule.limit
        remaining = max(0, rule.limit - count)

        # Calculate retry_after (time until next window)
        retry_after = rule.window - (current_time % rule.window)

        return allowed, remaining, retry_after



import functools
import inspect
from typing import Callable, Optional, TypeVar

from fastapi import Request

from rate_limiter.config import RateLimiterConfig
from rate_limiter.core.identity import IdentityResolver, get_identity_resolver
from rate_limiter.core.limiter import RateLimiter
from rate_limiter.core.rules import RateLimitRule
from rate_limiter.exceptions import RateLimitExceeded

F = TypeVar("F", bound=Callable)


def rate_limit(
    rule_str: str,
    key: str = "ip",
    config: Optional[RateLimiterConfig] = None,
    identity_resolver: Optional[IdentityResolver] = None,
) -> Callable[[F], F]:
    """
    Decorator for rate limiting FastAPI endpoints.

    Args:
        rule_str: Rate limit rule string (e.g., "5/minute")
        key: Identity key type ("ip", "user_id", "api_key")
        config: Optional rate limiter configuration
        identity_resolver: Optional custom identity resolver

    Returns:
        Decorated function

    Example:
        @app.get("/login")
        @rate_limit("5/minute", key="ip")
        async def login(request: Request):
            return {"message": "Login"}
    """
    rule = RateLimitRule.from_string(rule_str, key=key)
    limiter_config = config or RateLimiterConfig.from_env()
    limiter = RateLimiter(config=limiter_config, rules=[rule])
    resolver = identity_resolver or get_identity_resolver(
        key, trust_proxy=limiter_config.trust_proxy_headers
    )

    def decorator(func: F) -> F:
        # Check if function already has Request parameter
        sig = inspect.signature(func)
        has_request_param = any(
            param.annotation == Request or param.name == "request"
            for param in sig.parameters.values()
        )

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            # Find Request object in args/kwargs or get from FastAPI context
            request = None
            
            # Try to find in args/kwargs first
            for arg in args:
                if isinstance(arg, Request):
                    request = arg
                    break
            if request is None:
                request = kwargs.get("request")
            
            # If not found and function doesn't have request param, inject it
            if request is None and not has_request_param:
                # Try to get from FastAPI's dependency context
                try:
                    from starlette.requests import Request as StarletteRequest
                    # Check if we can get it from the current context
                    # This is a fallback - ideally the endpoint should have Request param
                    for arg in args:
                        if isinstance(arg, StarletteRequest):
                            request = arg
                            break
                except Exception:
                    pass
            
            # If still not found, try to inject Request dependency
            if request is None:
                # FastAPI should inject Request automatically, but if it's not in signature,
                # we need to get it from the context. For now, raise a helpful error.
                raise ValueError(
                    "Request object not found. Add 'request: Request' parameter to your endpoint function."
                )

            # Resolve identity
            identity = await resolver.resolve(request)

            # Get endpoint path
            endpoint = request.url.path if hasattr(request, "url") else ""

            # Check rate limits and get info in one call
            rate_limit_info, rate_limit_exception = await limiter.check_with_info(
                identity=identity, endpoint=endpoint
            )

            if rate_limit_exception:
                # Return 429 Too Many Requests with rate limit headers
                e = rate_limit_exception
                import time
                from fastapi import HTTPException

                reset_timestamp = int(time.time()) + e.retry_after
                headers = {
                    "Retry-After": str(e.retry_after),
                    "X-RateLimit-Limit": str(e.limit),
                    "X-RateLimit-Remaining": "0",
                    "X-RateLimit-Reset": str(reset_timestamp),
                }
                raise HTTPException(
                    status_code=429,
                    detail=f"Rate limit exceeded. Retry after {e.retry_after} seconds.",
                    headers=headers,
                )

            # Call the original function
            result = await func(*args, **kwargs)

            # Add rate limit headers to response
            if rate_limit_info:
                from fastapi import Response
                from starlette.responses import JSONResponse

                headers = rate_limit_info.to_headers()

                # If result is already a Response object, add headers to it
                if isinstance(result, Response):
                    for key, value in headers.items():
                        result.headers[key] = value
                    return result
                else:
                    # Wrap non-Response return values in JSONResponse with headers
                    # FastAPI will use this Response object instead of auto-wrapping
                    return JSONResponse(content=result, headers=headers)

            return result

        return wrapper  # type: ignore

    return decorator



"""Storage backends for rate limiting."""

from rate_limiter.core.storage.backend import StorageBackend
from rate_limiter.core.storage.memory import MemoryStorageBackend
from rate_limiter.core.storage.redis import RedisStorageBackend

__all__ = ["StorageBackend", "RedisStorageBackend", "MemoryStorageBackend"]
